#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "config.h"
#include "types.h"
#include "nlllib.h"
#include "label.h"

static struct label labels[LABEL_MAXNUM];

static label_t head = NULL;
static label_t pool = NULL;

int label_check(void)
{
  label_t label;
  int n = 0;

  if (head)
    return NLL_ERRCODE_LABEL_NOT_EMPTY;

  for (label = pool; label && (n < LABEL_MAXNUM); label = label->next)
    n++;

  if (label || (n != LABEL_MAXNUM))
    return NLL_ERRCODE_LABEL_NOT_EMPTY;

  return 0;
}

static int label_alloc(label_t *labelp)
{
  label_t label;

  if (!pool)
    return NLL_ERRCODE_LABEL_BUFFER_OVER;

  label = pool;
  pool = pool->next;

  memset(label, 0, sizeof(*label));

  *labelp = label;

  return 0;
}

static int label_free(label_t label)
{
  label_t next;

  for (; label; label = next) {
    next = label->next;

    label->next = pool;
    pool = label;
  }

  return 0;
}

int label_init(void)
{
  label_t label;
  int i;

  head = NULL;
  pool = NULL;
  memset(labels, 0, sizeof(labels));

  for (i = 0; i < LABEL_MAXNUM; i++) {
    label = &labels[i];
    label->next = pool;
    pool = label;
  }

  return 0;
}

int label_get(const char *name, label_t *labelp)
{
  label_t label;

  for (label = head; label; label = label->next) {
    if (!strcmp(label->name, name)) {
      *labelp = label;
      return 0;
    }
  }

  return NLL_ERRCODE_LABEL_NOT_FOUND;
}

int label_set(const char *name, spot_t spot, label_t *labelp)
{
  label_t label;
  int r;

  if (strlen(name) > LABEL_NAMELEN)
    return NLL_ERRCODE_LABEL_TOO_LONG;

  if (label_get(name, &label) == 0) {
    if (spot && label->spot.line)
      return NLL_ERRCODE_LABEL_DUPLICATE;
  } else {
    if ((r = label_alloc(&label)) < 0)
      return r;

    strcpy(label->name, name);

    label->next = head;
    head = label;
  }

  if (spot) {
    label->spot.line    = spot->line;
    label->spot.command = spot->command;
  }

  label->refcount++;

  *labelp = label;

  return 0;
}

int label_clear(label_t label)
{
  label->spot.line    = NULL;
  label->spot.command = NULL;
  return 0;
}

int label_clean(label_t label)
{
  return 0;
}

int label_del(label_t label)
{
  label_t *p, next;

  for (p = &head; *p;) {
    if (*p == label) {
      if (--(label->refcount) == 0) {
	next = (*p)->next;
	(*p)->next = NULL;
	label_free(*p);
	*p = next;
	continue;
      }
    }
    p = &((*p)->next);
  }

  return 0;
}

int label_dump(FILE *fp, label_t label)
{
  if (!label)
    label = head;

  nll_wait_output(fp);
  fprintf(fp, "\n-- Label dump --\n");

  for (; label; label = label->next) {
    nll_wait_output(fp);
    fprintf(fp, ".%s\tRef:%d\n", label->name, label->refcount);
  }

  fflush(fp);

  return 0;
}
