#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <sys/types.h>

#include "config.h"
#include "picture.h"

struct bmp_magic_header {
  uint8_t type[2];
};
#define BMP_MAGIC_HEADER_TYPE0   0
#define BMP_MAGIC_HEADER_TYPE1   1
#define BMP_MAGIC_HEADER_ALLSIZE 2

struct bmp_file_header {
  uint32_t size;
  uint16_t reserve[2];
  uint32_t offset;
};
#define BMP_FILE_HEADER_SIZE     0
#define BMP_FILE_HEADER_OFFSET   8
#define BMP_FILE_HEADER_ALLSIZE 12

struct bmp_info_header {
  uint32_t ihdrsize;
  uint32_t width;
  uint32_t height;
  uint16_t planes;
  uint16_t bits;
  uint32_t compress;
  uint32_t imagesize;
  uint32_t xpix;
  uint32_t ypix;
  uint32_t colors;
  uint32_t impcolors;
};
#define BMP_INFO_HEADER_IHDRSIZE   0
#define BMP_INFO_HEADER_WIDTH      4
#define BMP_INFO_HEADER_HEIGHT     8
#define BMP_INFO_HEADER_PLANES    12
#define BMP_INFO_HEADER_BITS      14
#define BMP_INFO_HEADER_COMPRESS  16
#define BMP_INFO_HEADER_IMAGESIZE 20
#define BMP_INFO_HEADER_XPIX      24
#define BMP_INFO_HEADER_YPIX      28
#define BMP_INFO_HEADER_COLORS    32
#define BMP_INFO_HEADER_IMPCOLORS 36
#define BMP_INFO_HEADER_ALLSIZE   40

#define BMP_PIXEL_BYTE 3

union evalue {
  uint8_t  b[4];
  uint16_t h[2];
  uint32_t w[1];
};

static uint16_t read_le16(uint16_t val)
{
  union evalue e;
  e.h[0] = val;
  return (e.b[1] << 8) | e.b[0];
}

static uint32_t read_le32(uint32_t val)
{
  union evalue e;
  e.w[0] = val;
  return (e.b[3] << 24) | (e.b[2] << 16) | (e.b[1] << 8) | e.b[0];
}

static uint16_t write_le16(uint16_t val)
{
  union evalue e;
  e.b[0] =  val        & 0xff;
  e.b[1] = (val >>  8) & 0xff;
  return e.h[0];
}

static uint32_t write_le32(uint32_t val)
{
  union evalue e;
  e.b[0] =  val        & 0xff;
  e.b[1] = (val >>  8) & 0xff;
  e.b[2] = (val >> 16) & 0xff;
  e.b[3] = (val >> 24) & 0xff;
  return e.w[0];
}

static int *palette_load(FILE *fp, int palette_num, int *palette)
{
  int i, s;
  unsigned char bgr[4];
  int c;

  for (i = 0; i < palette_num; i++) {
    s = fread(bgr, 1, sizeof(bgr), fp);
    if (s != sizeof(bgr))
      goto err;
    c  =  bgr[0];
    c |= (bgr[1] <<  8);
    c |= (bgr[2] << 16);
    palette[i] = c;
  }

  return palette;

err:
  return NULL;
}

static int *bmp_load(FILE *fp, int *widthp, int *heightp, int *palette,
		     int width, int height, int bits, int imagesize)
{
  int size, bytes, w, h, x, y, i, n, s, c, align_width;
  int *bitmap = NULL;
  unsigned char bgr[4], mask;

  w = (width  > GRAPHIC_MAXWIDTH ) ? GRAPHIC_MAXWIDTH  : width;
  h = (height > GRAPHIC_MAXHEIGHT) ? GRAPHIC_MAXHEIGHT : height;

  if ((w < 1) || (h < 1))
    goto err;

  size = w * h * sizeof(*bitmap);

  bitmap = malloc(size);
  if (!bitmap)
    goto err;
  memset(bitmap, 0, size);

  if (bits <= 8) {
    n = 8 / bits;
    bytes = 1;
    mask = (1 << bits) - 1;
  } else {
    n = 1;
    bytes = bits / 8;
    mask = 0xff;
  }

  size = 0;

  align_width = ((width + 3) / 4) * 4;

  for (y = 0; y < h; y++) {
    for (x = 0; x < align_width;) {
      for (i = 0; i < n; i++) {
	if (i == 0) {
	  if (size + bytes > imagesize)
	    goto err;
	  s = fread(bgr, 1, bytes, fp);
	  if (s < bytes)
	    memset(bgr + s, 0, bytes - s);
	  size += bytes;
	}
	if (palette) {
	  c = palette[(bgr[0] >> ((n - 1 - i) * bits)) & mask];
	} else {
	  c  =  bgr[0];
	  c |= (bgr[1] <<  8);
	  c |= (bgr[2] << 16);
	}
	if (x < w)
	  bitmap[(h - y - 1) * w + x] = c;
	x++;
      }
    }
  }

  if (widthp ) *widthp  = w;
  if (heightp) *heightp = h;

  return bitmap;

err:
  if (bitmap)
    free(bitmap);
  return NULL;
}

int *picture_bmp_load(char *filename, int *widthp, int *heightp)
{
  FILE *fp = NULL;
  struct bmp_magic_header mhdr;
  struct bmp_file_header fhdr;
  struct bmp_info_header ihdr;
  unsigned char dummy;
  int offset, n;
  int palette[256];
  int *p = NULL;
  int *bitmap;

  fp = fopen(filename, "r");
  if (!fp)
    goto err;

  if (fread(&mhdr, 1, BMP_MAGIC_HEADER_ALLSIZE, fp) != BMP_MAGIC_HEADER_ALLSIZE)
    goto err;

  offset = BMP_MAGIC_HEADER_ALLSIZE;

  if ((mhdr.type[0] != 'B') ||
      (mhdr.type[1] != 'M'))
    goto err;

  if (fread(&fhdr, 1, BMP_FILE_HEADER_ALLSIZE, fp) != BMP_FILE_HEADER_ALLSIZE)
    goto err;

  if (fread(&ihdr, 1, BMP_INFO_HEADER_ALLSIZE, fp) != BMP_INFO_HEADER_ALLSIZE)
    goto err;

  fhdr.size   = read_le32(fhdr.size);
  fhdr.offset = read_le32(fhdr.offset);

  offset += BMP_FILE_HEADER_ALLSIZE;

  ihdr.ihdrsize  = read_le32(ihdr.ihdrsize);
  ihdr.width     = read_le32(ihdr.width);
  ihdr.height    = read_le32(ihdr.height);
  ihdr.planes    = read_le16(ihdr.planes);
  ihdr.bits      = read_le16(ihdr.bits);
  ihdr.compress  = read_le32(ihdr.compress);
  ihdr.imagesize = read_le32(ihdr.imagesize);
  ihdr.xpix      = read_le32(ihdr.xpix);
  ihdr.ypix      = read_le32(ihdr.ypix);
  ihdr.colors    = read_le32(ihdr.colors);
  ihdr.impcolors = read_le32(ihdr.impcolors);

#if 0
  if (ihdr.ihdrsize != BMP_INFO_HEADER_ALLSIZE)
    goto err;
#endif
  if (ihdr.planes != 1)
    goto err;
  if (ihdr.compress != 0)
    goto err;

  if ((ihdr.bits != 1) && (ihdr.bits != 4) && (ihdr.bits != 8) &&
      (ihdr.bits != 24) && (ihdr.bits != 32))
    goto err;

  offset += BMP_INFO_HEADER_ALLSIZE;

  if (ihdr.bits <= 8) {
    n = ihdr.colors ? ihdr.colors : (1 << ihdr.bits);
    p = palette_load(fp, n, palette);
    if (!p)
      goto err;
    offset += n * sizeof(*p);
  }

  if (fhdr.offset < offset)
    goto err;

  for (offset = fhdr.offset - offset; offset > 0; offset--) {
    if (fread(&dummy, 1, 1, fp) != 1)
      goto err;
  }

  bitmap = bmp_load(fp, widthp, heightp, p, ihdr.width, ihdr.height, ihdr.bits, ihdr.imagesize);
  if (!bitmap)
    goto err;

  fclose(fp);

  return bitmap;

err:
  if (fp)
    fclose(fp);
  return NULL;
}

int picture_bmp_save(char *filename, int *bitmap, int bitmap_width, int bitmap_height,
		     int x, int y, int width, int height)
{
  FILE *fp = NULL;
  struct bmp_magic_header mhdr;
  struct bmp_file_header fhdr;
  struct bmp_info_header ihdr;
  int hdrsize, imgsize, align_width;
  int x0, y0, bx, by, c;
  unsigned char bgr[3];

  x0 = x;
  y0 = y;

  if (!width ) width  = bitmap_width  - x0;
  if (!height) height = bitmap_height - y0;

  fp = fopen(filename, "w");
  if (!fp)
    goto err;

  memset(&mhdr, 0, sizeof(mhdr));
  memset(&fhdr, 0, sizeof(fhdr));
  memset(&ihdr, 0, sizeof(ihdr));

  mhdr.type[0] = 'B';
  mhdr.type[1] = 'M';

  align_width = ((width + 3) / 4) * 4;

  hdrsize = BMP_MAGIC_HEADER_ALLSIZE + BMP_FILE_HEADER_ALLSIZE + BMP_INFO_HEADER_ALLSIZE;
  imgsize = align_width * height * sizeof(bgr);

  fhdr.size   = write_le32(hdrsize + imgsize);
  fhdr.offset = write_le32(hdrsize);

  ihdr.ihdrsize  = write_le32(BMP_INFO_HEADER_ALLSIZE);
  ihdr.width     = write_le32(width);
  ihdr.height    = write_le32(height);
  ihdr.planes    = write_le16(1);
  ihdr.bits      = write_le16(sizeof(bgr) * 8);
  ihdr.compress  = write_le32(0);
  ihdr.imagesize = write_le32(imgsize);
  ihdr.xpix      = write_le32(2925);
  ihdr.ypix      = write_le32(2925);
  ihdr.colors    = write_le32(0);
  ihdr.impcolors = write_le32(0);

  if (fwrite(&mhdr, 1, BMP_MAGIC_HEADER_ALLSIZE, fp) != BMP_MAGIC_HEADER_ALLSIZE)
    goto err;

  if (fwrite(&fhdr, 1, BMP_FILE_HEADER_ALLSIZE, fp) != BMP_FILE_HEADER_ALLSIZE)
    goto err;

  if (fwrite(&ihdr, 1, BMP_INFO_HEADER_ALLSIZE, fp) != BMP_INFO_HEADER_ALLSIZE)
    goto err;

  for (y = 0; y < height; y++) {
    for (x = 0; x < align_width; x++) {
      if (x < width) {
	bx = x0 + x;
	by = y0 + (height - y - 1);
	if ((bx < 0) || (bx > bitmap_width  - 1) ||
	    (by < 0) || (by > bitmap_height - 1))
	  c = 0;
	else
	  c = bitmap[by * bitmap_width + bx];
      } else {
	c = 0;
      }
      bgr[0] =  c        & 0xff;
      bgr[1] = (c >>  8) & 0xff;
      bgr[2] = (c >> 16) & 0xff;
      fwrite(bgr, 1, sizeof(bgr), fp);
    }
  }

  fclose(fp);

  return 0;

err:
  if (fp)
    fclose(fp);
  return -1;
}

int *picture_load(char *filename, int *widthp, int *heightp)
{
  return picture_bmp_load(filename, widthp, heightp);
}

int picture_save(char *filename, int *bitmap, int bitmap_width, int bitmap_height,
		 int x, int y, int width, int height)
{
  return picture_bmp_save(filename, bitmap, bitmap_width, bitmap_height,
			  x, y, width, height);
}
